/***************************************************************************
 *                                                                         *
 *   KNetLoad is copyright (c) 1999-2000, Markus Gustavsson                *
 *                         (c) 2002, Ben Burton                            *
 *                         (c) 2004-2005, Diego Pettenò                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "knetproc.h"
#include "settings.h"
#include "statgraph.h"
#include "statpopup.h"

#include <kiconloader.h>
#include <kglobal.h>
#include <qpixmap.h>
#include <klocale.h>
#include <kdebug.h>

// BSD-specific includes.
#ifdef Q_OS_BSD4
# include <stdlib.h>
# include <sys/param.h>
# include <sys/sysctl.h>
# include <sys/socket.h>
# include <net/if_dl.h>
# include <net/route.h>
# include <net/if.h>
#endif

// Solaris-specific includes.
#ifdef Q_OS_SOLARIS
# include <kstat.h>
# nclude <sys/sockio.h>
#endif

// Linux-specific includes.
#ifdef Q_OS_LINUX
# include <qdir.h>
# include <qregexp.h>
# include <qfile.h>
#endif

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifdef HAVE_STDINT_H
# include <stdint.h>
#elif defined HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

QStringList KNetProc::s_interfaces;

#define MAX_NET_DEV_LINE 512

KNetProc::KNetProc(QWidget *parent, uint index, const char *name) :
	QObject(parent, name), m_sets(NULL),
	currIn(0), currOut(0), prevIn(0), prevOut(0)
{
	kdDebug() << "KNetProc::KNetProc(" << parent << ", " << index << ", " << name << ")" << endl;
	
// OS-specific initialisation.
#ifdef Q_OS_BSD4
	buf = 0;
	alloc = 0;
#endif

	m_popup = new StatPopup(parent, "stat_popup");

	m_sets = new Settings(QString::number(index));
	m_sets->readConfig();
	
	if ( m_sets->type() == Settings::EnumType::Local )
	{
#ifdef Q_OS_LINUX
		rx.setName("/sys/class/net/" + m_sets->name() + "/statistics/rx_bytes");
		tx.setName("/sys/class/net/" + m_sets->name() + "/statistics/tx_bytes");
		useSys = QDir::home().exists("/sys/class/net");
#endif
	}

	m_in = new StatGraph(parent, 1, 0, "IN", StatGraph::defaultBgColor, "labelIn");
	connect(m_in, SIGNAL(clickedLeft()), this, SLOT(togglePopup()));
	m_out = new StatGraph(parent, 1, 0, "OUT", StatGraph::defaultBgColor, "labelOut");
	connect(m_out, SIGNAL(clickedLeft()), this, SLOT(togglePopup()));
	
	reloadSettings();
}

KNetProc::~KNetProc()
{
	kdDebug() << "KNetProc::~KNetProc()" << endl;
#ifdef Q_OS_BSD4
	free(buf);
#endif
	delete m_in;
	delete m_out;
}

void KNetProc::togglePopup()
{
	kdDebug() << "KNetProc::togglePopup()" << endl;
	if ( m_popup->isShown() )
		m_popup->hide();
	else
		m_popup->show();
}

void KNetProc::reloadSettings()
{
	kdDebug() << "KNetProc::reloadSettings()" << endl;
	m_sets->readConfig();
	
	m_in->setReadingColor(m_sets->color1());
	m_in->setStyle((StatGraph::Style)m_sets->style());
	m_in->setGridPace( m_sets->grid() ? 6 : 0 );
	m_in->setBgColor( m_sets->transparentBackground() ? StatGraph::invalidColor : m_sets->colorBackground() );
	m_in->setLabel( m_sets->labelled() ? "IN" : QString::null );
	m_in->update();

	m_out->setReadingColor(m_sets->color2());
	m_out->setStyle((StatGraph::Style)m_sets->style());
	m_out->setGridPace( m_sets->grid() ? 6 : 0 );
	m_out->setBgColor( m_sets->transparentBackground() ? StatGraph::invalidColor : m_sets->colorBackground() );
	m_out->setLabel( m_sets->labelled() ? "OUT" : QString::null );
	m_out->update();
	
	if ( ! m_sets->enabled() )
		m_popup->setText(i18n("Disabled."));
}

void KNetProc::takeReading()
{
	kdDebug() << "KNetProc::takeReading()" << endl;
	if ( ! m_sets->enabled() ) return;
	
	// The current readings must now become the previous readings.
	prevIn = currIn;
	prevOut = currOut;

	// Take the new set of readings (i.e., update currIn and currOut).
	
	// If no new readings are taken (e.g., the device could not be
	// found), this code may return; otherwise it *must* fall through to
	// the tidy-up block that follows this OS-specific code.
	readLoadActual();

	// Tidy up.
	if (currIn < prevIn)
		prevIn = 0;
	if (currOut < prevOut)
		prevIn = 0;
	
	float bitRateIn = recentBytesIn() * 8.0 / (m_sets->timeout() / 1000.0);
	float bitRateOut = recentBytesOut() * 8.0 / (m_sets->timeout() / 1000.0);
	
	// Convert to percentages.
	int8_t upperIn = (int8_t) (100.0 * bitRateIn / m_sets->scaleIn());
	if (upperIn < 0) upperIn = 0;
	else if (upperIn > 100) upperIn = 100;
	
	int8_t upperOut = (int8_t) (100.0 * bitRateOut /
	    (m_sets->scaleOut() ? m_sets->scaleOut() : m_sets->scaleIn() ) );
	if (upperOut < 0)
		upperOut = 0;
	else if (upperOut > 100)
		upperOut = 100;
	
	m_in->addPercentReading(upperIn, m_sets->soft());
	m_out->addPercentReading(upperOut, m_sets->soft());

	if ( !m_popup->isVisible() ) return;
	
	if ( m_sets->speedUnit() == 0 ) // bits per second
	{
		float currBitIn = (int)((bitRateIn + 50) / 100) / 10.0;
		float currBitOut = (int)((bitRateOut + 50) / 100) / 10.0;
		m_popup->setText(i18n(
			"Current In: %1 Kb/s, Total In: %2 MB.\n"
			"Current Out: %3 Kb/s, Total Out: %4 MB.")
			.arg(currBitIn).arg(totalMbIn())
			.arg(currBitOut).arg(totalMbOut()));
	} else { // bytes per second
		float currByteIn = (int)((bitRateIn/8 + 50) / 100) / 10.0;
		float currByteOut = (int)((bitRateOut/8 + 50) / 100) / 10.0;
		m_popup->setText(i18n(
			"Current In: %1 KB/s, Total In: %2 MB.\n"
			"Current Out: %3 KB/s, Total Out: %4 MB.")
			.arg(currByteIn).arg(totalMbIn())
			.arg(currByteOut).arg(totalMbOut()));
	}
	m_popup->resizeToText();
}

#ifdef Q_OS_LINUX
void KNetProc::readLoadActual()
{
	kdDebug() << "KNetProc::readLoadActual() [Linux]" << endl;
	if ( useSys )
		readLoadSys();
	else
		readLoadProc();
}

inline void KNetProc::readLoadSys()
{
	kdDebug() << "KNetProc::readLoadSys() [Linux]" << endl;
	QString read;
	
	if ( ! rx.open(IO_ReadOnly) )
		return;
	
	if ( rx.readLine(read, 32) == -1 )
	{
		rx.close();
		return;
	}
	
	read.stripWhiteSpace();
	currIn = read.toUInt();
	
	rx.close();
	
	if ( ! tx.open(IO_ReadOnly) )
		return;
	
	if ( tx.readLine(read, 32) == -1 )
	{
		tx.close();
		return;
	}
	
	read.stripWhiteSpace();
	currOut = read.toUInt();
	
	tx.close();
}

inline void KNetProc::readLoadProc()
{
	kdDebug() << "KNetProc::readLoadProc() [Linux]" << endl;
	static QFile proc("/proc/net/dev");
	static QRegExp procre("^\\s+(\\S+):(\\d+)\\s+(\\d+\\s+){7}(\\d+\\s+)(\\d+\\s*){7}$");
	QString read;
	
	if ( ! proc.open(IO_ReadOnly) )
		return;
	
	while ( ! proc.atEnd() )
	{
		if ( proc.readLine(read, MAX_NET_DEV_LINE) == -1 )
		{
			proc.close();
			return;
		}
		
		if ( procre.search(read) == -1 )
			continue;
		
		if ( procre.cap(1) != m_sets->name() )
			continue;
		
		currIn = procre.cap(2).toUInt();
		currOut = procre.cap(4).toUInt();
		break;
	}
	
	proc.close();
}
#endif

#ifdef Q_OS_BSD4
void KNetProc::readLoadActual() {
    kdDebug() << "KNetProc::readLoadActual() [BSD4]" << endl;
    static struct if_msghdr *ifm, *nextifm;
    static struct sockaddr_dl *sdl;
    static char *lim, *next;
    static size_t needed;
    static char s[32];
    static int mib[] = { CTL_NET, PF_ROUTE, 0, 0, NET_RT_IFLIST, 0 };
    if (sysctl(mib, 6, NULL, &needed, NULL, 0) < 0)
        return;
    if (alloc < needed) {
        if (buf != NULL)
            free(buf);
        buf = (char *) malloc(needed);
        if (buf == NULL)
            return;
        alloc = needed;
    }

    if (sysctl(mib, 6, buf, &needed, NULL, 0) < 0)
        return;
    lim = buf + needed;

    next = buf;
    while (next < lim) {
        ifm = (struct if_msghdr *) next;
        if (ifm->ifm_type != RTM_IFINFO)
            return;
        next += ifm->ifm_msglen;

        while (next < lim) {
            nextifm = (struct if_msghdr *) next;
            if (nextifm->ifm_type != RTM_NEWADDR)
                break;
            next += nextifm->ifm_msglen;
        }

        if (ifm->ifm_flags & IFF_UP) {
            sdl = (struct sockaddr_dl *) (ifm + 1);
            if (sdl->sdl_family != AF_LINK)
                continue;
            strncpy(s, sdl->sdl_data, sdl->sdl_nlen);
            s[sdl->sdl_nlen] = 0;

	    if (m_sets->name() == s) {
                currIn = ifm->ifm_data.ifi_ibytes;
                currOut = ifm->ifm_data.ifi_obytes;
                break;
            }
        }
    }
}
#endif

#ifdef Q_OS_SOLARIS
void KNetProc::readLoadActual() {
    kdDebug() << "KNetProc::readLoadActual() [Solaris]" << endl;
    static kstat_ctl_t *kc;
    static kstat_t *ksp;
    static kstat_named_t *knp;
    kc = kstat_open();
    ksp = kstat_lookup(kc, 0, -1, m_sets->name());
    if (ksp && kstat_read(kc, ksp, 0) >= 0) {
        if ((knp = (kstat_named_t *) kstat_data_lookup(ksp, "rbytes")))
            currIn = knp->value.ui32;
        if ((knp = (kstat_named_t *) kstat_data_lookup(ksp, "obytes")))
            currOut = knp->value.ui32;
    }
    kstat_close(kc);
}
#endif

const QStringList &KNetProc::loadInterfaces(bool force)
{
	kdDebug() << "KNetProc::loadInterface(" << force << ")" << endl;
	if ( ! force && ! s_interfaces.empty() )
		return s_interfaces;
	
	if ( force )
		s_interfaces.clear();

#ifndef Q_OS_LINUX
	s_interfaces.push_back("lo");
	s_interfaces.push_back("eth0");
	s_interfaces.push_back("ppp0");
	s_interfaces.push_back("ippp0");
#else
	if ( QDir::root().exists("/sys/class/net") )
	{ // Exists /sys, 2.6 series kernel
		QDir sys("/sys/class/net");
		QStringList l = sys.entryList();
		QStringList::const_iterator end = l.end();
		for(QStringList::const_iterator it = l.begin(); it != end; it++)
		{
			if ( (*it)[0] == '.' ) continue;
			s_interfaces.push_back(*it);
		}
	} else { // Doesn't exists, kernel 2.4 or earlier
		static QFile proc("/proc/net/dev");
		static QRegExp procre("^\\s+(\\S+):(\\d+\\s*){17}$");
		QString read;
		
		if ( ! proc.open(IO_ReadOnly) )
			return s_interfaces;
		
		while ( ! proc.atEnd() )
		{
			if ( proc.readLine(read, MAX_NET_DEV_LINE) == -1 )
			{
				proc.close();
				return s_interfaces;
			}
			
			if ( procre.search(read) == -1 )
				continue;
			
			s_interfaces.push_back(procre.cap(1));
		}
		
		proc.close();
	}
#endif
	return s_interfaces;
}

const QPixmap KNetProc::getInterfaceIcon(const QString &iface)
{
	kdDebug() << "KNetProc::getInterfaceIcon(" << iface << ")" << endl;
	
	QString icon = "devother";
	if ( iface == "lo" )
		icon = "devlo";
	else if ( iface.find("eth") == 0 )
		icon = "deveth0";
	else if ( iface.find("ppp") == 0 )
		icon = "devppp0";
	else if ( iface.find("ippp") == 0 )
		icon = "devippp0";
	
	return KGlobal::iconLoader()->loadIcon(icon, KIcon::Small);
}

#include "knetproc.moc"
