/***************************************************************************
 *                                                                         *
 *   KNetLoad is copyright (c) 1999-2000, Markus Gustavsson                *
 *                         (c) 2002, Ben Burton                            *
 *                         (c) 2004-2005, Diego Pettenò                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef __KNETPROC_H
#define __KNETPROC_H

#include <qglobal.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qpixmap.h>
#include <qobject.h>

#ifdef Q_OS_LINUX
# include <qfile.h>
#endif

static const float megabyte = 1024*1024;

class StatGraph;
class StatPopup;
class Settings;

/**
 * A class used to read raw network throughput details from the system.
 *
 * Linux
 * -----
 *
 * For kernels 2.4 and earlier, we use /proc/net/dev file, which contains
 * lines structured in this way:
 * 
 *   iface: bytesIn x x x x x x x bytesOut x x x x x x x
 * 
 * The first two lines are headers, but also if they are not, there are
 * no problems: the regular expression make simpler to parse it.
 * 
 * For kernels 2.5 and greater, if /sys directory is present, it's used
 * to get the statistics.
 *
 * BSD
 * ---
 *
 * BSD code by Roland Riegel <feedback@roland-riegel.de>,
 * Bill Wilson <bill@gkrellm.net>, Hajimu Umemoto <ume@mahoroba.org> and
 * Anthony Mallet <anthony.mallet@useless-ficus.net>.
 *
 * Solaris
 * -------
 *
 * Solaris code by Roland Riegel <feedback@roland-riegel.de>,
 * Bill Wilson <bill@gkrellm.net> and Daisuke Yabuki <dxy@acm.org>.
 *
 * Other Operating Systems
 * -----------------------
 *
 * Please, send in a patch!
 *
 * The KNetLoad maintainer is currently Diego Pettenò
 * <flameeyes@users.berlios.de>, or you could submit a patch through the KDE
 * bug tracking system ( http://bugs.kde.org/ ).
 */
class KNetProc : public QObject {
Q_OBJECT
public:
	/**
	 * @brief Constructor
	 *
	 * @param in Input load graph
	 * @param out Output load graph
	 * @param device Interface to read the load for
	 * @param parent Parent Object
	 * @param name Name of the instance
	 */
	KNetProc(QWidget *parent, uint index, const char *name = "");
	~KNetProc();

	/**
	 * @brief Loads the list of interfaces, and return them
	 * @param force Force reload also if m_interfaces is empty
	 * @todo On operating systems different from Linux it fills it with
	 *       stock devices which can not be present in the system.
	 */
	static const QStringList &loadInterfaces(bool force = false);
	
	/**
	 * @brief Gets the icon for the requested interface
	 * @param iface Interface to get the icon for
	 */
	static const QPixmap getInterfaceIcon(const QString &iface);
	
	StatGraph *graphIn()
	{ return m_in; }
	
	StatGraph *graphOut()
	{ return m_out; }

public slots:
	/**
	 * @brief Reloads the settings
	 *
	 * This function takes care of reloading the settings from the config file,
	 * changing the graphs when needed.
	 */
	void reloadSettings();
	
	/**
	 * @brief Takes the reading
	 *
	 * This function is used to avoid code redundancy in KNetLoad application
	 * and applet.
	 */
	void takeReading();
	
	/**
	 * @brief Toggle showing of the popup
	 *
	 * This function make the popup show (if hidden) or hide (if shown)
	 */
	void togglePopup();

private: 
//@{   
/**
 * @name dev Device and results of readings.
 */
	static QStringList s_interfaces;
		///< List of all the interfaces found in the system
	
	Settings *m_sets;
		///< Settings instance for the device
	uint currIn, currOut;
		///< Total traffic in bytes from the last reading.
	uint prevIn, prevOut;
		///< Total traffic in bytes from the next-to-last reading.
	StatGraph *m_in, *m_out;
		///< Graphs for input and output of the interface
	StatPopup *m_popup;
		///< Popup with statistics of the interface
	
	void readLoadActual();
		///< Do the actual reading
//@}

	/**
	 * The total number of bytes that have passed through the selected
	 * device between the last reading and the previous reading.
	 *
	 * The values returned are guaranteed to be non-negative.
	 */
	uint recentBytesIn() const
	{ return currIn - prevIn; }
	
	uint recentBytesOut() const
	{ return currOut - prevOut; }

	/**
	 * The total amount of traffic that has passed through the selected
	 * device, measured in megabytes.  This amount is measured over the
	 * entire history of the device, not just from the last reading.
	 *
	 * See the const KNetProc::megabyte for exactly what constitutes a
	 * megabyte for our purposes.
	 */
	inline float totalMbIn() const
	{ return currIn / megabyte; }
	
	inline float totalMbOut() const
	{ return currOut / megabyte; }


//@{
/**
 * @name osspec OS-specific data members.
 */
#ifdef Q_OS_BSD4
	char* buf;
		///< A temporary buffer whose size may need to change.
	size_t alloc;
		///< The number of bytes currently allocated for buf.
#endif
#ifdef Q_OS_LINUX
	bool useSys;
		///< If true, the /sys interface will be used.
	inline void readLoadProc();
		///< Read the load from the old /proc interface (2.4 and earlier).
	inline void readLoadSys();
		///< Read the load from the new /sys interface (since 2.5).
	
	QFile rx;
		///< File object for /sys receiving load.
	QFile tx;
		///< File object for /proc transmitted load.
#endif
//@}
};

#endif
